/*
 * afcopy.cpp:
 *
 * Copy one AFF file to another. 
 * Resulting file is re-ordered and possibly re-compressed.
 */

/*
 * Copyright (c) 2006
 *	Simson L. Garfinkel 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. [Omitted]
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL SIMSON
 * GARFINKEL, BAIS TECHNOLOGy, OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "afflib.h"
#include "afflib_i.h"
#include "quads.h"

#include <ctype.h>
#include <zlib.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <assert.h>

#include <algorithm>
#include <cstdlib>
#include <vector>
#include <string>
#include <map>

using namespace std;

#ifdef UNIX
#include <unistd.h>
#include <term.h>
#define HAS_CURSES
#endif

#ifdef WIN32
#include "unix4win32.h"
#include <malloc.h>
#endif

char *progname = "afcopy";

#define xstr(s) str(s)
#define str(s) #s

int opt_all = 0;
int opt_debug = 0;
int opt_verbose = 0;
int opt_x = 0;
int opt_y = 0;
int opt_preen = 0;

void usage()
{
    printf("%s version %s\n",progname,xstr(AFFLIB_VERSION));
    printf("usage: %s [options] file1 file\n",progname);
    printf("                    Copies file1 to file2\n");
    printf("       %s [options] file1 file2 file3 ... dir\n",progname);
    printf("                    Copies file1.. into dir\n");
    printf("       %s [options] file1 file2 file3 ... dir1 dir2...\n",progname);
    printf("                    Copies file1.. into dirs1, dir2, ...\n");
    printf("\n");
    printf("By default, all page MACs are verified on read and all segments\n");
    printf("are verified after write.\n");
    
    printf("Options:\n");
    printf("   -a = print each file as it is copied\n");
    printf("   -d = print each segment as it is copied (debug)\n");
    printf("   -c = also verify decompression of each page\n");
    printf("   -x = don't verify hashes on reads\n");
    printf("   -y = don't verify writes\n");
    printf("   -p = preen; recompress all pages with LZMA or NULL\n");
    printf("\n");
    printf("   -h = help; print this message.\n");
    printf("   -v = print the program version and exit.\n");
    exit(1);
}


const char *current_source = 0;
const char *current_dest = 0;
const char *current_seg  = 0;
void sig_info(int arg)
{
    if(!current_source || !current_dest) return;
    printf("Copying %s --> %s",current_source,current_dest);
    if(current_seg) printf(" (%s) ",current_seg);
    printf("\n");
}


/* Structure for map */
struct less_c_str
{
    inline bool operator()( const char* x, const char* y) const
    {     return ( strcmp( x,y ) < 0 );
    }
};

struct md5blob {
    unsigned char buf[16];
};

typedef map< const char*, struct md5blob, less_c_str > hashMapT;


/* Copy pagenumber from ain to aout. Return 0 if success, -1 if can't do it. */
int preen(AFFILE *ain,AFFILE *aout,int64 pagenum,unsigned long arg)
{
    int alg = (arg & AF_PAGE_COMP_ALG_MASK);
    //int max = (alg & AF_PAGE_COMP_MAX);

    if(alg==AF_PAGE_COMP_ALG_ZERO) return -1; // can't preen ZERO
    if(alg==AF_PAGE_COMP_ALG_LZMA) return -1; // don't re-compress LZMA
    
    size_t pagesize = af_page_size(ain);
    if(pagesize<=0) return -1;		// couldn't get pagesize

    unsigned char *pagebuf = (unsigned char *)malloc(pagesize);
    if(!pagebuf) return -1;		// couldn't allocate memory for page?

    if(af_get_page(ain,pagenum,pagebuf,&pagesize)){
	free(pagebuf);
	return -1;
    }

    /* Got the page; now write it out with max compression */
    af_enable_compression(aout,AF_COMPRESSION_ALG_LZMA,AF_COMPRESSION_MAX);

    /* Write out the page */
    if(af_update_page(aout,pagenum,pagebuf,pagesize)){
	free(pagebuf);
	return -1;
    }
    free(pagebuf);
    return 0;
}

#ifndef __BSD_VISIBLE
#define isnumber(x) isdigit(x)
#endif

int afcopy(char *infile,vector<string> &outfiles)
{
#ifdef SIGINFO
    signal(SIGINFO,sig_info);
#endif
    hashMapT hashMap;
    
    AFFILE *ain = af_open(infile,O_RDONLY,0);
    if(!ain) err(1,infile);

    vector<AFFILE *>afouts;				    // vector of output AFFs
    vector<int64>preened_pages;
    current_source = infile;
    
    if(opt_all) printf("%s: ",af_filename(ain));
    for(vector<string>::const_iterator o = outfiles.begin();
	o != outfiles.end();
	o++){
	const char *fname = o->c_str();
	if(access(fname,F_OK)==0){
	    fprintf(stderr,"%s: file exists\n",fname);
	  continue;
	}
	
	AFFILE *aout = af_open(fname,O_RDWR|O_CREAT|O_TRUNC,0666);
	if(aout==0){
	  warn(fname);
	  continue;
	}
	if(o != outfiles.begin()) printf("\t ");
	printf(" => %s\n",fname);
	if(af_set_pagesize(aout,af_page_size(ain))){
	    errx(1,"%s: cannot set page size to %d\n",
		 af_filename(aout),af_page_size(ain));
	}
	afouts.push_back(aout);
    }
    if(afouts.size()==0) return -1;


    /* Get a list of all the segments */
    af_rewind_seg(ain);
    char segname[AF_MAX_NAME_LEN];
    vector <string> segments;
    memset(segname,0,sizeof(segname));
    while(af_get_next_seg(ain,segname,sizeof(segname),0,0,0)==0){
	segments.push_back(segname);
    }


    for(vector<string>::const_iterator seg = segments.begin(); seg!= segments.end();seg++){
	/* For each segment, get the size of the segment */

	const char *segname = seg->c_str();
	current_seg = segname;

	size_t seglen=0;
	if(af_get_seg(ain,segname,0,0,&seglen)){
	    err(1,"Cannot read length of segment '%s' in %s",segname,af_filename(ain));
	}
	unsigned char *segbuf = (unsigned char *)malloc(seglen);
	if(!segbuf){
	    err(1,"Cannot allocated %d bytes for segment '%s' in %s",
		(int)seglen,segname,af_filename(ain));
	}

	/* Now get the segment */
	unsigned long arg;
	if(af_get_seg(ain,segname,&arg,segbuf,&seglen)){
	    err(1,"Cannot read segment '%s' in %s",segname,af_filename(ain));
	}
	int64 pagenumber = af_segname_page_number(segname);

	/* Calculate the MD5 of this segment and remember it in the map */
	md5blob md5;
	MD5(segbuf,seglen,md5.buf);
	hashMap[segname] = md5;


	/* Write the segment to each file */
	for(vector<AFFILE *>::iterator aout = afouts.begin();
	    aout != afouts.end();
	    aout++){

	    current_dest = af_filename(*aout);
	    if(opt_debug) printf("   %s -> %s ...",segname,af_filename(*aout));

	    if(pagenumber>=0 && opt_preen){
		if(opt_debug) printf(" (PREENED) ");
		if(preen(ain,*aout,pagenumber,arg)==0){
		    preened_pages.push_back(pagenumber);
		    continue;
		}
	    }
	    if(af_update_seg(*aout,segname,arg,segbuf,seglen,1)){
		err(1,"Cannot write segment '%s' to %s",segname,af_filename(*aout));
	    }
	    if(opt_debug) printf("\n");
	}
	free(segbuf);
	current_dest = 0;
    }
    current_seg = 0;
    af_close(ain);

    current_seg = "VERIFYING";
    /* Now verify all of the hashes */
    if(opt_debug) printf("\n\nFiles copied. Verifying...\n");
    int badcopy = 0;
    for(vector<string>::const_iterator seg = segments.begin(); seg!= segments.end();seg++){
	const char *segname = seg->c_str();
	for(vector<AFFILE *>::iterator aout = afouts.begin();
	    aout != afouts.end();
	    aout++){
	    size_t seglen=0;
	    char b2[1024];
	again:
	    if(af_get_seg(*aout,segname,0,0,&seglen)){
		if(segname != b2 &&
		   segname[0]=='s' && segname[1]=='e' && segname[2]=='g' &&
		   isnumber(segname[3])){
		    /* Looks like a legacy segname name was renamed.
		     * Try the new name
		     */
		    sprintf(b2,"page%s",segname+3);
		    segname = b2;
		    goto again;
		}
		warn("Cannot read length of segment '%s' in output file %s",
		     segname,af_filename(*aout));
		badcopy = 1;
		continue;
	    }
	    int64 pagenumber = af_segname_page_number(segname);
	    if(find(preened_pages.begin(),preened_pages.end(),pagenumber)!=preened_pages.end()){
		/* TK: page pagenumber was preened. It should be check against the original hash */
		continue;
	    }
	    

	    unsigned char *segbuf = (unsigned char *)malloc(seglen);
	    if(!segbuf){
		warn("Cannot allocated %d bytes for segment '%s' in %s", (int)seglen,segname,af_filename(ain));
		badcopy = 1;
		continue;
	    }
	    unsigned long arg;
	    if(af_get_seg(*aout,segname,&arg,segbuf,&seglen)){
		warn("Cannot read segment '%s' in %s",segname,af_filename(*aout));
		free(segbuf);
		badcopy = 1;
		continue;
	    }

	    /* Calculate the MD5 of this segment and see if it matches the map.
	     * (But don't do this for preened segments.
	     */
	    unsigned char md5_read[16];
	    MD5(segbuf,seglen,md5_read);
	    if(memcmp(hashMap[segname].buf,md5_read,16)!=0){
		warn("Hash read from %s for segment %s doesn't validate.", segname,af_filename(*aout));
		badcopy = 1;
	    }
	    free(segbuf);		// free the buffer
	}
    }
    for(vector<AFFILE *>::iterator aout = afouts.begin(); aout != afouts.end(); aout++){
	af_close(*aout);
    }
    if(opt_debug) printf("\n");
    current_source = 0;
    return 0;
}

int main(int argc,char **argv)
{
    int ch;

    setvbuf(stdout,0,_IONBF,0);		// turn off buffering on stdout
    while ((ch = getopt(argc, argv, "adpvxyh?")) != -1) {
	switch (ch) {
	case 'a': opt_all++; break;
	case 'd': opt_debug++; break;
	case 'p': opt_preen++;break;
	case 'x': opt_x++;break;
	case 'y': opt_y++;break;

	case 'h':
	case '?':
	default:
	    usage();
	    break;
	case 'v':
	    printf("%s version %s\n",progname,xstr(AFFLIB_VERSION));
	    exit(0);
	}
    }
    argc -= optind;
    argv += optind;

    if(argc<2){				// at this point, we need at least two args
	usage();
    }


    /* Find any directories */
    vector<string> dirlist;
    for(int i=argc-1;i>0;i--){
	struct stat st;
	
	if(stat(argv[i],&st)!=0) break; // out of directories
	if((st.st_mode & S_IFMT)!=S_IFDIR) break; // found a non-dir
	dirlist.push_back(argv[i]);
	argc--;			// ignore the last
    }

    /* If I found no directories, then there better just be two values */
    

    if(dirlist.size()==0){
	if(argc!=2){
	    fprintf(stderr,"Please specify a directory or just two AFF files.\n\n");
	    usage();
	}
	/* Must be copying from file1 to file2. Make sure file2 does not exist */
	if(access(argv[1],R_OK)==0){
	    errx(1,"File exists: %s\n",argv[1]);
	}
	
	vector<string> outfiles;
	outfiles.push_back(argv[1]);
	return afcopy(argv[0],outfiles);
    }

    /* Loop for each file and each directory */

    while(argc--){
	/* Open the output files */
	vector<string> outfiles;
	for(unsigned int i=0;i<dirlist.size();i++){
	    char outfilename[MAXPATHLEN+1];
	    const char *name = rindex(*argv,'/');
	    if(name) name++;
	    else name = *argv;
	    sprintf(outfilename,"%s/%s",dirlist[i].c_str(),name);
	    outfiles.push_back(outfilename);
	}
	afcopy(argv[0],outfiles);	   // old outfiles will get GCed
	argv++;
    }
    return 0;
}


